import tempfile
import os

from argparsetree import BaseCommand

from ..validation.model_data import csv_validity_test
from ..utils.path import as_path
from ..utils.inputs import InputValues

from argparse import RawDescriptionHelpFormatter
from .command import OasisBaseCommand, OasisComputationCommand



class ModelValidationCmd(OasisBaseCommand):
    """
    Checks the validity of a set of model data.
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """

        super(self.__class__, self).add_args(parser)

        parser.add_argument(
            '-d', '--model-data-path',
            default=None, help='Directory containing additional user-supplied model data files')

    def action(self, args):
        """
        Performs validity checks on model data csv files using ktools
        executables.

        :param args: The arguments from the command line
        :type args: Namespace
        """

        inputs = InputValues(args)

        model_data_fp = as_path(
            inputs.get('model_data_path', required=True, is_path=True), 'Model data path', is_dir=True)

        csv_validity_test(model_data_fp)


class FmValidationCmd(OasisComputationCommand):
    """
    Runs a set of FM tests.
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'RunFmTest'


class GenerateDummyModelFilesCmd(OasisComputationCommand):
    
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'GenerateDummyModelFiles'


class GenerateDummyOasisFilesCmd(OasisComputationCommand):

    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'GenerateDummyOasisFiles'


class GenerateLossesDummyModelCmd(OasisComputationCommand):

    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'GenerateLossesDummyModel'


class TestModelCmd(BaseCommand):
    """
    Present sub-commands for creating test models
    """

    sub_commands = {
        'generate-model-files': GenerateDummyModelFilesCmd,
        'generate-oasis-files': GenerateDummyOasisFilesCmd,
        'run': GenerateLossesDummyModelCmd
    }


class TestCmd(BaseCommand):
    """
    Test models and keys servers
    """

    sub_commands = {
        'model-validation': ModelValidationCmd,
        'fm': FmValidationCmd,
        'model': TestModelCmd
    }



__version__ = '1.16.0'

import sys
import os
from importlib.abc import MetaPathFinder, Loader
import importlib
import warnings

import logging
from logging import NullHandler

logger = logging.getLogger(__name__)
logger.addHandler(NullHandler())


class MyLoader(Loader):

    def __init__(self, sub_module):
        self.sub_module = sub_module

    def module_repr(self, module):
        return repr(module)

    def load_module(self, fullname):
        old_name = fullname
        names = fullname.split(".")
        names[1] = self.sub_module
        fullname = ".".join(names)
        module = importlib.import_module(fullname)
        sys.modules[old_name] = module
        return module


class MyImport(MetaPathFinder):
    """ Support alias of depreciated sub-modules

        * model_execution   -> execution
        * model_preparation -> preparation
        * api               -> platform

        Example:
            `from oasislmf.model_execution.bash import genbash`
                is the same as calling the new name
            `from oasislmf.execution.bash import genbash`
    """

    def __init__(self):
        self.depricated_modules = {
            "model_execution": "execution",
            "model_preparation": "preparation",
            "api": "platform",
        }

    def find_module(self, fullname, path=None):
        names = fullname.split(".")
        if len(names) >= 2:
            if names[0] == "oasislmf" and names[1] in self.depricated_modules:
                deprecated = "imports from 'oasislmf.{}' are deprecated. Import by using 'oasislmf.{}' instead.".format(
                    names[1],
                    self.depricated_modules[names[1]]
                )
                warnings.simplefilter("always")
                warnings.warn(deprecated)
                return MyLoader(self.depricated_modules[names[1]])

sys.meta_path.append(MyImport())



__all__ = [
    'OasisBuiltinBaseLookup',
    'OasisBaseKeysLookup',
    'OasisLookup',
    'OasisPerilLookup',
    'OasisVulnerabilityLookup',
    'OasisLookupFactory'
]


## This is a backwards compatibility file for lookup code based on version 1.9.0 and older
## Add deprecated warning, Load equiv classes from ..lookup. * 

from ..lookup.factory import KeyServerFactory as OasisLookupFactory
from ..lookup.base import OasisBaseLookup as OasisBuiltinBaseLookup
from ..lookup.interface import OasisLookupInterface as OasisBaseKeysLookup 
from ..lookup.rtree import RTreeLookup as OasisLookup
from ..lookup.rtree import RTreePerilLookup as OasisPerilLookup
from ..lookup.rtree import RTreeVulnerabilityLookup as OasisVulnerabilityLookup



__all__ = [
    'KeyServerInterface',
    'OasisLookupInterface',
]

# 'OasisBaseKeysLookup' -> OasisLookupInterface


import os
import abc

from ..utils.log import oasis_log
from ..utils.status import OASIS_KEYS_STATUS


''' Interface class for developing custom key server or key lookup code
'''


class KeyServerInterface(metaclass=abc.ABCMeta):
    """
    Interface to implement to create a KeyServer
    It define the method to be implemented to be used correctly in lookup.factory.KeyServerFactory
    all classes must:
     - specify the version of the interface they use
     - implement the init method
     - implement the generate_key_files method

    """
    interface_version = "1"

    @abc.abstractmethod
    def __init__(self, config, config_dir, user_data_dir, output_dir):
        """
        During the key generation step, the generic factory will call the constructor of the lookup class with the
        following parameters.

        :param config: contains all the information necessary to run the model
        :type config: dict

        :param config_dir: path to the model directory, can be used to locate relative path to all the files
                           that serve as base for the model
        :type config_dir: str

        :param user_data_dir: Path to additional data necessary for the model that can vary from analysis to analysis
        :type user_data_dir: str

        :param output_dir: Path to the analysis output directory, can be use to write additional files that are produce
                           during the keys file generation

        """

        raise NotImplementedError

    @abc.abstractmethod
    def generate_key_files(self,
                           location_fp,
                           successes_fp,
                           errors_fp=None,
                           output_format='oasis',
                           keys_success_msg=False,
                           multiproc_enabled=True,
                           multiproc_num_cores=-1,
                           multiproc_num_partitions=-1,
                           **kwargs):
        """
        Writes a keys file, and optionally a keys error file.

        :param location_fp: path to the locations file
        :type location_fp: str

        :param successes_fp: path to the success keys file
        :type successes_fp: str

        :param errors_fp: path to the error keys file (optional)
        :type errors_fp: str

        :param output_format: format of the keys files (oasis or json)
        :type output_format: str

        :param keys_success_msg: option to write msg for success key
        :type keys_success_msg: bool

        :param multiproc_enabled: option to run with multiple processor
        :type multiproc_enabled: bool

        :param multiproc_num_cores: number of cores to use in multiproc mode
        :type multiproc_num_cores: int

        :param multiproc_num_partitions: number of partition to create in multiproc mode
        :type multiproc_num_partitions: int

        If ``keys_errors_file_path`` is not present then the method returns a
        pair ``(p, n)`` where ``p`` is the keys file path and ``n`` is the
        number of "successful" keys records written to the keys file, otherwise
        it returns a quadruple ``(p1, n1, p2, n2)`` where ``p1`` is the keys
        file path, ``n1`` is the number of "successful" keys records written to
        the keys file, ``p2`` is the keys errors file path and ``n2`` is the
        number of "unsuccessful" keys records written to keys errors file.
        """
        raise NotImplementedError


class KeyLookupInterface(metaclass=abc.ABCMeta):
    """Interface for KeyLookup
    it define the interface to be used correctly by lookup.factory.BasicKeyServer
    all classes must:
     - specify the version of the interface they use
     - implement the init method
     - implement the process_location method
    """

    interface_version = "1"

    @abc.abstractmethod
    def __init__(self, config, config_dir, user_data_dir, output_dir):
        raise NotImplementedError

    @abc.abstractmethod
    def process_locations(self, loc_df):
        """
        Process location rows - passed in as a pandas dataframe.
        """
        raise NotImplementedError


class OasisLookupInterface:  # pragma: no cover
    """
    Old Oasis base class -deprecated
    If you were using this interface, you can make you class inherit from the new abstract class AbstractBasicKeyServer
    or implement the KeyServerInterface interface
    """
    interface_version = "0"

    @oasis_log()
    def __init__(
        self,
        keys_data_directory=None,
        supplier=None,
        model_name=None,
        model_version=None,
        complex_lookup_config_fp=None,
        output_directory=None
    ):
        """
        Class constructor
        """
        if keys_data_directory is not None:
            self.keys_data_directory = keys_data_directory
        else:
            self.keys_data_directory = os.path.join(os.sep, 'var', 'oasis', 'keys_data')

        self.supplier = supplier
        self.model_name = model_name
        self.model_version = model_version
        self.complex_lookup_config_fp = complex_lookup_config_fp
        self.output_directory = output_directory
        self.UNKNOWN_ID = -1

    @oasis_log()
    def process_locations(self, loc_df):
        """
        Process location rows - passed in as a pandas dataframe.
        """
        pass

    def _get_area_peril_id(self, record):
        """
        Get the area peril ID for a particular location record.
        """
        return self.UNKNOWN_ID, "Not implemented"

    def _get_vulnerability_id(self, record):
        """
        Get the vulnerability ID for a particular location record.
        """
        return self.UNKNOWN_ID, "Not implemented"

    @oasis_log()
    def _get_area_peril_ids(self, loc_data, include_context=True):
        """
        Generates area peril IDs in two modes - if include_context is
        True (default) it will generate location records/rows including
        the area peril IDs, otherwise it will generate pairs of location
        IDs and the corresponding area peril IDs.
        """
        pass

    @oasis_log()
    def _get_vulnerability_ids(self, loc_data, include_context=True):
        """
        Generates vulnerability IDs in two modes - if include_context is
        True (default) it will generate location records/rows including
        the area peril IDs, otherwise it will generate pairs of location
        IDs and the corresponding vulnerability IDs.
        """
        pass

    def _get_custom_lookup_success(self, ap_id, vul_id):
        """
        Determine the status of the keys lookup.
        """
        if ap_id == self.UNKNOWN_ID or vul_id == self.UNKNOWN_ID:
            return OASIS_KEYS_STATUS['nomatch']['id']
        return OASIS_KEYS_STATUS['success']['id']



__all__ = [
    'ExposurePreAnalysis'
]

import json
import os
import pathlib
import shutil

from ..base import ComputationStep
from ...utils.path import get_custom_module
from ...utils.exceptions import OasisException
from ...utils.defaults import store_exposure_fp


class ExposurePreAnalysis(ComputationStep):
    """
    Computation step that will be call before the gulcalc.
    Add the ability to specify a model specific pre-analysis hook for exposure modification,
    Allows OED to be processed by some custom code.
    Example of usage include geo-coding, exposure enhancement, or dis-aggregation...

    when the run method is call it will :
    - load the module specified at exposure_pre_analysis_module
    - init the class named exposure_pre_analysis_class_name with all the non null args in step_params as key arguments
    - call the method run of the object
    - return the output of the method

    you can find an example of such custom module in OasisPyWind/custom_module/exposure_pre_analysis.py

    """
    step_params = [{'name': 'exposure_pre_analysis_module', 'required': True, 'is_path': True, 'pre_exist': True,
                    'help': 'Exposure Pre-Analysis lookup module path'},
                   {'name': 'exposure_pre_analysis_class_name', 'default': 'ExposurePreAnalysis',
                    'help': 'Name of the class to use for the exposure_pre_analysis'},
                   {'name': 'exposure_pre_analysis_setting_json', 'is_path': True, 'pre_exist': True,
                    'help': 'Exposure Pre-Analysis config JSON file path'},
                   {'name': 'oed_location_csv', 'is_path': True, 'pre_exist': True,
                    'help': 'Source location CSV file path'},
                   {'name': 'oed_accounts_csv', 'is_path': True, 'pre_exist': True,
                    'help': 'Source accounts CSV file path'},
                   {'name': 'oed_info_csv', 'is_path': True, 'pre_exist': True,
                    'help': 'Reinsurance info. CSV file path'},
                   {'name': 'oed_scope_csv', 'is_path': True, 'pre_exist': True,
                    'help': 'Reinsurance scope CSV file path'},
                   {'name': 'oasis_files_dir', 'flag': '-o', 'is_path': True, 'pre_exist': False,
                    'help': 'Path to the directory in which to generate the Oasis files'},
                   ]

    run_dir_key = 'pre-analysis'

    def run(self):
        """
        import exposure_pre_analysis_module and call the run method
        """

        # If given a value for 'oasis_files_dir' then use that directly
        if self.oasis_files_dir:
            input_dir = self.oasis_files_dir
        else:
            input_dir = self.get_default_run_dir()
            pathlib.Path(input_dir).mkdir(parents=True, exist_ok=True)

        kwargs = {}
        for input_name in ('oed_location_csv', 'oed_accounts_csv', 'oed_info_csv', 'oed_scope_csv'):
            file_path_in = getattr(self, input_name)
            if file_path_in is not None:
                file_path_raw = os.path.join(input_dir, f'epa_{store_exposure_fp(file_path_in, input_name)}')
                file_path_out = os.path.join(input_dir, store_exposure_fp(file_path_in, input_name))
                kwargs[f'raw_{input_name}'] = file_path_raw
                kwargs[input_name] = file_path_out

                shutil.copyfile(file_path_in, file_path_raw)
                shutil.copyfile(file_path_in, file_path_out)

        if self.exposure_pre_analysis_setting_json:
            with open(self.exposure_pre_analysis_setting_json) as exposure_pre_analysis_setting_file:
                kwargs['exposure_pre_analysis_setting'] = json.load(exposure_pre_analysis_setting_file)

        _module = get_custom_module(self.exposure_pre_analysis_module, 'Exposure Pre-Analysis lookup module path')

        try:
            _class = getattr(_module, self.exposure_pre_analysis_class_name)
        except AttributeError as e:
            raise OasisException(f"class {self.exposure_pre_analysis_class_name} "
                                 f"is not defined in module {self.exposure_pre_analysis_module}") from e.__cause__

        self.logger.info('\nPre-analysis modified files: {}'.format(
            json.dumps({k:v for (k,v) in kwargs.items() if k.startswith('oed')}, indent=4))
        )
        self.logger.info('\nPre-analysis original files: {}'.format(
            json.dumps({k:v for (k,v) in kwargs.items() if k.startswith('raw')}, indent=4))
        )
        return _class(**kwargs).run()



__all__ = [
    'csv_validity_test',
]

import os
import logging
import subprocess

from ..utils.exceptions import OasisException

# Model files to test
INPUT_FILES = {
    'damage_bin_dict': {
        'name': 'damage_bin_dict',
        'validation_tool': 'validatedamagebin',
        'flag': '-d'
    },
    'footprint': {
        'name': 'footprint',
        'validation_tool': 'validatefootprint',
        'flag': '-f'
    },
    'vulnerability': {
        'name': 'vulnerability',
        'validation_tool': 'validatevulnerability',
        'flag': '-s'
    }
}


def csv_validity_test(model_data_fp):
    """
    Assess validity of model data.

    :param model_data_fp: directory containing csv files
    :type model_data_fp: str

    :raises OasisException: if one of the tests fail
    """

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    model_data_dir = os.path.abspath(model_data_fp)

    # Check individual files
    for input_file in INPUT_FILES.values():
        validation_tool = input_file['validation_tool']
        input_file_path = os.path.join(
            model_data_dir,
            '{}.csv'.format(input_file['name'])
        )

        logger.info("Testing {}.csv".format(input_file['name']))
        cmd_str = "{} < {}".format(validation_tool, input_file_path)

        try:
            subprocess.check_call(cmd_str, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as e:
            raise OasisException("Exception raised in 'csv_validity_test'", e)

    # Execute cross checks
    logger.info("Executing cross checks")
    cmd_str = "crossvalidation"
    for input_file in INPUT_FILES.values():
        flag = input_file['flag']
        input_file_path = os.path.join(
            model_data_dir,
            '{}.csv'.format(input_file['name'])
        )
        cmd_str += " {} {}".format(flag, input_file_path)

    try:
        subprocess.check_call(cmd_str, stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as e:
        raise OasisException("Exception raised in 'csv_validity_test'", e)



## TO DO -- placeholder for post analysis output modification hook 

#__all__ = [
#    '_'
#]
#
#
#class _(ComputationStep):
#    """
#    - desc here -   
#    """
#    step_params = [  .. args ..  ]
#
#
#    def run(self):
#        return None



__all__ = [
    'AbstractBasicKeyLookup',
    'MultiprocLookupMixin',
    'OasisBaseLookup',
]


# 'OasisBuiltinBaseLookup' -> 'OasisBaseLookup'

import abc
import itertools
import json
import os
import types

import pandas as pd

from ..utils.path import as_path
from ..utils.exceptions import OasisException

UNKNOWN_ID = -1


''' Basic abstract classes that facilitate the implementation of KeyLookupInterface  
'''


class AbstractBasicKeyLookup:
    """Basic abstract class for KeyLookup"""

    interface_version = "1"

    def __init__(self, config, config_dir=None, user_data_dir=None, output_dir=None):
        self.config = config
        self.config_dir = config_dir or '.'
        self.user_data_dir = user_data_dir
        self.output_dir = output_dir

    @abc.abstractmethod
    def process_locations(self, locations):
        """
        Process location rows - passed in as a pandas dataframe.
        Results can be list, tuple, generator or a pandas dataframe.
        """
        raise NotImplementedError


class MultiprocLookupMixin:
    """
    Simple mixin class for multiprocessing

    implement the process_locations_multiproc by transforming the result of process_locations into a pandas DataFrame
    """

    def process_locations_multiproc(self, loc_df_part):
        result = self.process_locations(loc_df_part)
        if isinstance(result, list) or isinstance(result, tuple):
            return pd.DataFrame(result)
        elif isinstance(result, types.GeneratorType):
            return pd.DataFrame.from_records(result)
        elif isinstance(result, pd.DataFrame):
            return result
        else:
            raise OasisException("Unrecognised type for results: {type(results)}. expected ")


class OasisBaseLookup(AbstractBasicKeyLookup, MultiprocLookupMixin):
    """
    Abstract class that help with the implementation of the KeyServerInterface.
    require lookup method to be implemented.
    Lookup will be call to create a key for each peril id and coverage type
    """
    multiproc_enabled = True

    def __init__(self, config=None, config_json=None, config_fp=None, config_dir=None, user_data_dir=None, output_dir=None):
        if config:
            config_dir = config_dir or '.'
        elif config_json:
            config = json.loads(config_json)
            config_dir = config_dir or '.'
        elif config_fp:
            config_dir = config_dir or os.path.dirname(config_fp)
            _config_fp = as_path(config_fp, 'config_fp')
            with open(_config_fp, 'r', encoding='utf-8') as f:
                config = json.load(f)

        keys_data_path = config.get('keys_data_path')
        keys_data_path = os.path.join(config_dir, keys_data_path) if keys_data_path else ''

        config['keys_data_path'] = as_path(keys_data_path, 'keys_data_path', preexists=(True if keys_data_path else False))

        super().__init__(config, config_dir=config_dir, user_data_dir=user_data_dir, output_dir=output_dir)

        peril_config = self.config.get('peril') or {}

        self.peril_ids = tuple(peril_config.get('peril_ids') or ())

        self.peril_id_col = peril_config.get('peril_id_col') or 'peril_id'

        coverage_config = self.config.get('coverage') or {}

        self.coverage_types = tuple(coverage_config.get('coverage_types') or ())

        self.coverage_type_col = peril_config.get('coverage_type_col') or 'coverage_type'

        self.config.setdefault('exposure', self.config.get('exposure') or self.config.get('locations') or {})

    def __tweak_config_data__(self):
        for section in ('exposure', 'peril', 'vulnerability',):
            section_config = self._config.get(section) or {}
            for k, v in section_config.items():
                if isinstance(v, str) and '%%KEYS_DATA_PATH%%' in v:
                    self._config[section][k] = v.replace('%%KEYS_DATA_PATH%%', self._config['keys_data_path'])
                elif type(v) == list:
                    self._config[section][k] = tuple(v)
                elif isinstance(v, dict):
                    for _k, _v in v.items():
                        if isinstance(_v, str) and '%%KEYS_DATA_PATH%%' in _v:
                            self._config[section][k][_k] = _v.replace('%%KEYS_DATA_PATH%%', self._config['keys_data_path'])

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, c):
        self._config = c
        self.__tweak_config_data__()

    def lookup(self, loc, peril_id, coverage_type, **kwargs):
        """
        Lookup for an individual location item, which could be a dict or a
        Pandas series object.
        """
        raise NotImplementedError

    def process_locations(self, locs):
        """
        Bulk vulnerability lookup for a list, tuple, generator, pandas data
        frame or dict of location items, which can be dicts or Pandas series
        objects or any object which has as a dict-like interface.

        Generates results using ``yield``.
        """
        locs_seq = None

        if isinstance(locs, list) or isinstance(locs, tuple):
            locs_seq = (loc for loc in locs)
        elif isinstance(locs, types.GeneratorType):
            locs_seq = locs
        elif isinstance(locs, dict):
            locs_seq = locs.values()
        elif isinstance(locs, pd.DataFrame):
            locs_seq = (loc for _, loc in locs.iterrows())

        for loc, peril_id, coverage_type in itertools.product(locs_seq, self.peril_ids, self.coverage_types):
            yield self.lookup(loc, peril_id, coverage_type)



__all__ = [
    'DeterministicLookup',
    'RTreeLookup',
    'RTreePerilLookup',
    'RTreeVulnerabilityLookup',
    'generate_index_entries',
    'get_peril_areas',
    'get_peril_areas_index',
    'get_rtree_index',
    'PerilArea',
    'PerilAreasIndex',
]

# 'OasisLookup' -> 'RTreeLookup'
# 'OasisPerilLookup' -> RTreePerilLookup
# 'OasisVulnerabilityLookup' -> 'RTreeVulnerabilityLookup'

from .base import OasisBaseLookup, AbstractBasicKeyLookup

import copy
import re
import types
import builtins
import itertools
import os
import uuid
import pickle
import pandas as pd

from collections import OrderedDict

from rtree.core import RTreeError
from rtree.index import (
    Index as RTreeIndex,
    Property as RTreeIndexProperty,
)

from shapely import speedups as shapely_speedups
from shapely.geometry import (
    box,
    Point,
    MultiPoint,
    Polygon,
)

from ..utils.data import get_dataframe
from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from ..utils.defaults import DEFAULT_RTREE_INDEX_PROPS
from ..utils.status import OASIS_KEYS_STATUS
from ..utils.path import as_path

if shapely_speedups.available:
    shapely_speedups.enable()


class DeterministicLookup(AbstractBasicKeyLookup):
    multiproc_enabled = False

    def process_locations(self, locations):
        loc_ids = (loc_it['loc_id'] for _, loc_it in locations.loc[:, ['loc_id']].sort_values('loc_id').iterrows())
        success_status= OASIS_KEYS_STATUS['success']['id']
        return pd.DataFrame.from_records((
            {'loc_id': _loc_id, 'peril_id': peril, 'coverage_type': cov_type, 'area_peril_id': i + 1,
             'vulnerability_id': i + 1, 'status': success_status}
            for i, (_loc_id, peril, cov_type) in enumerate(itertools.product(loc_ids, range(1, 1 + self.config['num_subperils']),
                                                                             self.config['supported_oed_coverage_types']))
        ))


# ---- RTree Lookup classes ---------------------------------------------------


class RTreeLookup(OasisBaseLookup):
    """
    Combined peril and vulnerability lookup
    """
    @oasis_log()
    def __init__(
        self,
        config=None,
        config_json=None,
        config_fp=None,
        config_dir=None,
        areas=None,
        peril_areas=None,
        peril_areas_index=None,
        peril_areas_index_props=None,
        loc_to_global_areas_boundary_min_distance=0,
        vulnerabilities=None,
        user_data_dir=None,
        output_dir=None
    ):
        super(self.__class__, self).__init__(
            config=config,
            config_json=config_json,
            config_fp=config_fp,
            config_dir=config_dir,
            user_data_dir=user_data_dir,
            output_dir=output_dir
        )

        self.peril_lookup = RTreePerilLookup(
            config=self.config,
            config_dir=self.config_dir,
            areas=areas,
            peril_areas=peril_areas,
            peril_areas_index=peril_areas_index,
            peril_areas_index_props=peril_areas_index_props,
            loc_to_global_areas_boundary_min_distance=loc_to_global_areas_boundary_min_distance
        )

        self.peril_area_id_key = str(str(self.config['peril'].get('peril_area_id_col') or '') or 'peril_area_id').lower()

        self.vulnerability_id_key = str(str(self.config['vulnerability'].get('vulnerability_id_col')) or 'vulnerability_id').lower()

        self.vulnerability_lookup = RTreeVulnerabilityLookup(
            config=self.config,
            config_dir=self.config_dir,
            vulnerabilities=vulnerabilities
        )

    def __new__(cls,
                config=None,
                config_json=None,
                config_fp=None,
                config_dir=None,
                areas=None,
                peril_areas=None,
                peril_areas_index=None,
                peril_areas_index_props=None,
                loc_to_global_areas_boundary_min_distance=0,
                vulnerabilities=None,
                user_data_dir=None,
                output_dir=None):
        if config:
            builtin_lookup_type = config.get('builtin_lookup_type', 'combined')
        else:
            builtin_lookup_type = 'combined'

        if builtin_lookup_type == 'combined':
            return super().__new__(cls)
        elif builtin_lookup_type == 'peril':
            return RTreePerilLookup(config=config,
                                    config_json=config_json,
                                    config_fp=config_fp,
                                    config_dir=config_dir,
                                    areas=areas,
                                    peril_areas=peril_areas,
                                    peril_areas_index=peril_areas_index,
                                    peril_areas_index_props=peril_areas_index_props,
                                    loc_to_global_areas_boundary_min_distance=loc_to_global_areas_boundary_min_distance,
                                    user_data_dir=user_data_dir,
                                    output_dir=output_dir)
        elif builtin_lookup_type == 'vulnerability':
            return RTreeVulnerabilityLookup(config=config,
                                            config_json=config_json,
                                            config_fp=config_fp,
                                            config_dir=config_dir,
                                            vulnerabilities=vulnerabilities,
                                            user_data_dir=user_data_dir,
                                            output_dir=output_dir)
        else:
            raise OasisException(f'Unknown builtin_lookup_type: {builtin_lookup_type}')

    def lookup(self, loc, peril_id, coverage_type, **kwargs):

        loc_id = loc.get('loc_id') or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        plookup = self.peril_lookup.lookup(loc, peril_id, coverage_type)

        past = plookup['status']
        pamsg = plookup['message']
        paid = plookup['peril_area_id']

        vlookup = self.vulnerability_lookup.lookup(loc, peril_id, coverage_type)

        vlnst = vlookup['status']
        vlnmsg = vlookup['message']
        vlnid = vlookup['vulnerability_id']

        vlookup.pop('status')
        vlookup.pop('message')
        vlookup.pop('vulnerability_id')

        # Could optionally call the status lookup method, but faster
        # to avoid or minimise outside function calls in a `for` loop
        status = (
            OASIS_KEYS_STATUS['success']['id'] if past == vlnst == OASIS_KEYS_STATUS['success']['id']
            else (OASIS_KEYS_STATUS['fail']['id'] if (past == OASIS_KEYS_STATUS['fail']['id'] or vlnst == OASIS_KEYS_STATUS['fail']['id']) else OASIS_KEYS_STATUS['nomatch']['id'])
        )

        message = '{}; {}'.format(pamsg, vlnmsg)

        return {
            k: v for k, v in itertools.chain(
                (
                    ('loc_id', loc_id),
                    ('peril_id', peril_id),
                    ('coverage_type', coverage_type),
                    (self.peril_area_id_key, paid),
                    (self.vulnerability_id_key, vlnid),
                    ('status', status),
                    ('message', message),
                ),
                vlookup.items()
            )
        }


class RTreePerilLookup(OasisBaseLookup):
    """
    Single peril, single coverage type, lon/lat point-area poly lookup using
    an Rtree index to store peril areas - index entries are
    #
    #    (peril area ID, peril area bounds)
    #
    # pairs. Areas must be represented as polygons with vertices which are
    # lon/lat coordinates, and be passed in to the constructor as a list,
    # tuple, or generator of triples of the form
    #
    #    (peril area ID, polygon lon/lat vertices, dict with optional properties)
    #
    # An optional distance measure ``loc_to_global_areas_boundary_min_distance`` can be passed
    # that defines how far, in an abstract unit, a given lon/lat location can be
    # from the boundary of the polygon containing all the individual peril area
    # polygons in order to be assigned an peril area ID. By default this distance
    # is 0, which means any lon/lat location outside the polygon containing all
    # peril area polygons will not be assigned a peril area ID.
    """

    @oasis_log()
    def __init__(
        self,
        areas=None,
        config=None,
        config_json=None,
        config_fp=None,
        config_dir=None,
        loc_to_global_areas_boundary_min_distance=0,
        peril_areas=None,
        peril_areas_index=None,
        peril_areas_index_fp=None,
        peril_areas_index_props=None,
        user_data_dir=None,
        output_dir=None
    ):
        super(self.__class__, self).__init__(
            config=config,
            config_json=config_json,
            config_fp=config_fp,
            config_dir=config_dir,
            user_data_dir=user_data_dir,
            output_dir=output_dir
        )

        peril_config = self.config.get('peril') or {}

        if areas or peril_areas or peril_config:
            if peril_areas_index:
                self.peril_areas_index = peril_areas_index
                self.peril_areas_index_props = self.peril_areas_index_props.properties.as_dict()
            elif (areas or peril_areas):
                self.index_props = (
                    peril_areas_index_props or
                    peril_config.get('rtree_index') or
                    DEFAULT_RTREE_INDEX_PROPS
                )
                self.peril_areas_index = PerilAreasIndex(areas=areas, peril_areas=peril_areas, properties=self.index_props)
            else:
                areas_rtree_index_config = peril_config.get('rtree_index') or {}
                index_fp = peril_areas_index_fp or areas_rtree_index_config.get('filename')

                if not os.path.isabs(index_fp):
                    index_fp = os.path.join(self.config_dir, index_fp)
                    index_fp = as_path(index_fp, 'index_fp', preexists=False)

                if index_fp:
                    idx_ext = areas_rtree_index_config.get('idx_extension') or 'idx'
                    dat_ext = areas_rtree_index_config.get('dat_extension') or 'dat'
                    if not (os.path.exists('{}.{}'.format(index_fp, idx_ext)) or os.path.exists('{}.{}'.format(index_fp, dat_ext))):
                        raise OasisException('No Rtree file index {}.{{{}, {}}} found'.format(index_fp, idx_ext, dat_ext))
                    self.peril_areas_index = PerilAreasIndex(fp=index_fp)
                    self.peril_areas_index_props = self.peril_areas_index.properties.as_dict()

            self.peril_areas_boundary = box(*self.peril_areas_index.bounds, ccw=False)

            _centroid = self.peril_areas_boundary.centroid
            self.peril_areas_centre = _centroid.x, _centroid.y

            self.loc_to_global_areas_boundary_min_distance = (
                loc_to_global_areas_boundary_min_distance or
                self.config['peril'].get('loc_to_global_areas_boundary_min_distance') or 0
            )

        if self.config.get('exposure') or self.config.get('locations'):
            self.loc_coords_x_col = str.lower(str(self.config['exposure'].get('coords_x_col')) or 'lon')
            self.loc_coords_y_col = str.lower(str(self.config['exposure'].get('coords_y_col')) or 'lat')
            self.loc_coords_x_bounds = tuple(self.config['exposure'].get('coords_x_bounds') or ()) or (-180, 180)
            self.loc_coords_y_bounds = tuple(self.config['exposure'].get('coords_y_bounds') or ()) or (-90, 90)

    def lookup(self, loc, peril_id, coverage_type, **kwargs):
        """
        Area peril lookup for an individual lon/lat location item, which can be
        provided as a dict or a Pandas series. The data structure should contain
        the keys `lon` or `longitude` for longitude and `lat` or `latitude` for
        latitude.
        """
        idx = self.peril_areas_index
        boundary = self.peril_areas_boundary
        loc_to_areas_min_dist = self.loc_to_global_areas_boundary_min_distance

        loc_id = loc.get('loc_id') or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        loc_x_col = self.loc_coords_x_col
        loc_y_col = self.loc_coords_y_col
        loc_x_bounds = self.loc_coords_x_bounds
        loc_y_bounds = self.loc_coords_y_bounds

        x = loc.get(loc_x_col)
        y = loc.get(loc_y_col)

        def _lookup(loc_id, x, y, st, perid, covtype, paid, pabnds, pacoords, msg):
            return {
                'loc_id': loc_id,
                loc_x_col: x,
                loc_y_col: y,
                'peril_id': perid,
                'coverage_type': covtype,
                'status': st,
                'peril_area_id': paid,
                'area_peril_id': paid,
                'area_bounds': pabnds,
                'area_coordinates': pacoords,
                'message': msg
            }

        try:
            x = float(x)
            y = float(y)
            if not ((loc_x_bounds[0] <= x <= loc_x_bounds[1]) and (loc_y_bounds[0] <= y <= loc_y_bounds[1])):
                raise ValueError('{}/{} out of bounds'.format(loc_x_col, loc_y_col))
        except (ValueError, TypeError) as e:
            msg = (
                'Peril area lookup: invalid {}/{} ({}, {}) - {}'
                .format(loc_x_col, loc_y_col, x, y, str(e))
            )
            return _lookup(loc_id, x, y, OASIS_KEYS_STATUS['fail']['id'], peril_id, coverage_type, None, None, None, msg)

        st = OASIS_KEYS_STATUS['nomatch']['id']
        msg = 'No peril area match'
        paid = None
        pabnds = None
        pacoords = None
        point = x, y

        try:
            results = list(idx.intersection(point, objects='raw'))

            if not results:
                raise IndexError

            for _perid, _covtype, _paid, _pabnds, _pacoords in results:
                if (peril_id, coverage_type) == (_perid, _covtype):
                    paid, pabnds, pacoords = _paid, _pabnds, _pacoords
                    break

            if paid is None:
                raise IndexError
        except IndexError:
            try:
                results = list(idx.nearest(point, objects='raw'))

                if not results:
                    raise IndexError

                for _perid, _covtype, _paid, _pabnds, _pacoords in results:
                    if (peril_id, coverage_type) == (_perid, _covtype):
                        paid, pabnds, pacoords = _paid, _pabnds, _pacoords
                        break

                if paid is None:
                    msg = 'No intersecting or nearest peril area found for peril ID {} and coverage type {}'.format(peril_id, coverage_type)
                    return _lookup(loc_id, x, y, OASIS_KEYS_STATUS['nomatch']['id'], peril_id, coverage_type, None, None, None, msg)
            except IndexError:
                pass
            else:
                p = Point(x, y)
                min_dist = p.distance(boundary)
                if min_dist > loc_to_areas_min_dist:
                    msg = (
                        'Peril area lookup: location is {} units from the '
                        'peril areas global boundary -  the required minimum '
                        'distance is {} units'
                        .format(min_dist, loc_to_areas_min_dist)
                    )
                    return _lookup(loc_id, x, y, OASIS_KEYS_STATUS['fail']['id'], peril_id, coverage_type, None, None, None, msg)
                st = OASIS_KEYS_STATUS['success']['id']
                msg = (
                    'Successful peril area lookup: {}'.format(paid)
                )
        except RTreeError as e:
            return _lookup(loc_id, x, y, OASIS_KEYS_STATUS['fail']['id'], peril_id, coverage_type, None, None, None, str(e))
        else:
            st = OASIS_KEYS_STATUS['success']['id']
            msg = 'Successful peril area lookup: {}'.format(paid)

        return _lookup(loc_id, x, y, st, peril_id, coverage_type, paid, pabnds, pacoords, msg)


class RTreeVulnerabilityLookup(OasisBaseLookup):
    """
    Simple key-value based vulnerability lookup
    """

    @oasis_log()
    def __init__(
        self,
        config=None,
        config_json=None,
        config_fp=None,
        config_dir=None,
        vulnerabilities=None,
        user_data_dir=None,
        output_dir=None
    ):
        super(self.__class__, self).__init__(
            config=config,
            config_json=config_json,
            config_fp=config_fp,
            config_dir=config_dir,
            user_data_dir=user_data_dir,
            output_dir=output_dir
        )

        if vulnerabilities or self.config.get('vulnerability'):
            self.col_dtypes, self.key_cols, self.vuln_id_col, self.vulnerabilities = self.get_vulnerabilities(vulnerabilities=vulnerabilities)

    @oasis_log()
    def get_vulnerabilities(self, vulnerabilities=None):
        if not self.config:
            raise OasisException(
                'No lookup configuration provided or set - use `get_config` '
                'on this instance to set it and provide either an actual '
                'model config dict (use `model_config` argument), or a model '
                'config JSON string (use `model_config_json` argument, or a '
                'model config JSON file path (use `model_config_fp` argument)'
            )

        vuln_config = self.config.get('vulnerability')

        if not vuln_config:
            raise OasisException('No vulnerability config set in the lookup config')

        col_dtypes = vuln_config.get('col_dtypes')

        if not col_dtypes:
            raise OasisException(
                'Vulnerability file column data types must be defined as a '
                '(col, data type) dict in the vulnerability section of the '
                'lookup config'
            )

        col_dtypes = {
            k.lower(): getattr(builtins, v) for k, v in col_dtypes.items()
        }

        key_cols = vuln_config.get('key_cols')

        if not vuln_config.get('key_cols'):
            raise OasisException(
                'The vulnerability file key column names must be listed in the '
                'vulnerability section of the lookup config'
            )

        key_cols = tuple(col.lower() for col in key_cols)

        vuln_id_col = str(str(self.config['vulnerability'].get('vulnerability_id_col')) or 'vulnerability_id').lower()

        def _vuln_dict(vulns_seq, key_cols, vuln_id_col):
            return (
                {v[key_cols[0]]: (v.get(vuln_id_col) or v.get('vulnerability_id')) for _, v in vulns_seq} if len(key_cols) == 1
                else OrderedDict(
                    {tuple(v[key_cols[i]] for i in range(len(key_cols))): (v.get(vuln_id_col) or v.get('vulnerability_id')) for v in vulns_seq}
                )
            )

        if vulnerabilities:
            return col_dtypes, key_cols, vuln_id_col, _vuln_dict(enumerate(vulnerabilities), key_cols)

        src_fp = vuln_config.get('file_path')

        if not src_fp:
            raise OasisException(
                'No vulnerabilities file path provided in the lookup config'
            )

        if not os.path.isabs(src_fp):
            src_fp = os.path.join(self.config_dir, src_fp)
            src_fp = os.path.abspath(src_fp)

        self.config['vulnerability']['file_path'] = src_fp

        src_type = str(str(vuln_config.get('file_type')) or 'csv').lower()

        float_precision = 'high' if vuln_config.get('float_precision_high') else None

        non_na_cols = vuln_config.get('non_na_cols') or tuple(col.lower() for col in list(key_cols) + [vuln_id_col])

        sort_cols = vuln_config.get('sort_cols') or vuln_id_col
        sort_ascending = vuln_config.get('sort_ascending')

        vuln_df = get_dataframe(
            src_fp=src_fp,
            src_type=src_type,
            float_precision=float_precision,
            lowercase_cols=True,
            non_na_cols=non_na_cols,
            col_dtypes=col_dtypes,
            sort_cols=sort_cols,
            sort_ascending=sort_ascending
        )

        return col_dtypes, key_cols, vuln_id_col, _vuln_dict((v for _, v in vuln_df.iterrows()), key_cols, vuln_id_col)

    def lookup(self, loc, peril_id, coverage_type, **kwargs):
        """
        Vulnerability lookup for an individual location item, which could be a dict or a
        Pandas series.
        """
        loc_id = loc.get('loc_id') or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        key_cols = self.key_cols
        col_dtypes = self.col_dtypes

        loc_key_col_values = OrderedDict({
            key_col: loc.get(key_col) for key_col in key_cols
        })

        if not loc_key_col_values['peril_id']:
            loc_key_col_values['peril_id'] = peril_id

        if not loc_key_col_values['coverage_type']:
            loc_key_col_values['coverage_type'] = loc.get('coverage') or coverage_type

        def _lookup(loc_id, vlnperid, vlncovtype, vlnst, vlnid, vlnmsg):
            return {
                k: v for k, v in itertools.chain(
                    (
                        ('loc_id', loc_id),
                        ('peril_id', vlnperid),
                        ('coverage_type', vlncovtype),
                        ('status', vlnst),
                        ('vulnerability_id', vlnid),
                        ('message', vlnmsg)
                    ),
                    loc_key_col_values.items()
                )
            }

        try:
            for key_col in key_cols:
                key_col_dtype = col_dtypes[key_col]
                key_col_dtype(loc_key_col_values[key_col])
        except (TypeError, ValueError):
            return _lookup(loc_id, peril_id, coverage_type, OASIS_KEYS_STATUS['fail']['id'], None, 'Vulnerability lookup: invalid key column value(s) for location')

        vlnperid = peril_id
        vlncovtype = coverage_type
        vlnst = OASIS_KEYS_STATUS['nomatch']['id']
        vlnmsg = 'No vulnerability match'
        vlnid = None

        try:
            vlnid = (
                self.vulnerabilities[tuple(loc_key_col_values[col] for col in key_cols)] if len(key_cols) > 1
                else self.vulnerabilities[loc[key_cols[0]]]
            )
        except KeyError:
            pass
        else:
            vlnperid = peril_id
            vlncovtype = coverage_type
            vlnst = OASIS_KEYS_STATUS['success']['id']
            vlnmsg = 'Successful vulnerability lookup: {}'.format(vlnid)

        return _lookup(loc_id, vlnperid, vlncovtype, vlnst, vlnid, vlnmsg)



# ---- RTree Peril funcs ------------------------------------------------------


def generate_index_entries(items, objects=None):
    if objects:
        for (key, poly_bounds), obj in zip(items, objects):
            yield key, poly_bounds, obj
    else:
        for key, poly_bounds in items:
            yield key, poly_bounds, None


def get_peril_areas(areas):
    for peril_id, coverage_type, peril_area_id, coordinates, other_props in areas:
        yield PerilArea(coordinates, peril_id=peril_id, coverage_type=coverage_type, peril_area_id=peril_area_id, **other_props)


def get_peril_areas_index(
    areas=None,
    peril_areas=None,
    properties=None
):
    if not (areas or peril_areas):
        raise OasisException('Either areas or peril areas must be provided')

    return PerilAreasIndex(areas=areas, peril_areas=peril_areas, properties=properties)


def get_rtree_index(
    items,
    objects=None,
    properties=None
):
    return (
        RTreeIndex(generate_index_entries(items, objects=objects), properties=RTreeIndexProperty(**properties)) if properties
        else RTreeIndex(generate_index_entries(items, objects=objects))
    )




class PerilArea(Polygon):

    def __init__(self, coords, **kwargs):

        _coords = tuple(c for c in coords)

        if not _coords:
            raise OasisException('No peril area coordinates')

        if len(_coords) > 2:
            self._multipoint = MultiPoint(_coords)
        elif len(_coords) == 2:
            minx, miny, maxx, maxy = tuple(_c for c in _coords for _c in c)
            self._multipoint = MultiPoint(box(minx, miny, maxx, maxy).exterior.coords)
        elif len(_coords) == 1:
            x, y = _coords[0][0], _coords[0][1]
            r = kwargs.get('area_reg_poly_radius') or 0.0016
            self._multipoint = MultiPoint(
                tuple((x + r * (-1)**i, y + r * (-1)**j) for i in range(2) for j in range(2))
            )

        super(self.__class__, self).__init__(shell=self._multipoint.convex_hull.exterior.coords)

        self._coordinates = tuple(self.exterior.coords)

        self._centre = self.centroid.x, self.centroid.y

        self._coverage_type = kwargs.get('coverage_type')

        self._peril_id = kwargs.get('peril_id')

        self._id = kwargs.get('area_peril_id') or kwargs.get('peril_area_id') or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

    @property
    def multipoint(self):
        return self._multipoint

    @property
    def coordinates(self):
        return self._coordinates

    @property
    def centre(self):
        return self._centre

    @property
    def coverage_type(self):
        return self._coverage_type

    @property
    def peril_id(self):
        return self._peril_id

    @property
    def id(self):
        return self._id


class PerilAreasIndex(RTreeIndex):

    def __init__(self, *args, **kwargs):

        self._protocol = pickle.HIGHEST_PROTOCOL

        idx_fp = kwargs.get('fp')

        areas = kwargs.get('areas')
        peril_areas = kwargs.get('peril_areas')

        props = kwargs.get('properties') or copy.deepcopy(DEFAULT_RTREE_INDEX_PROPS)

        if not (idx_fp or areas or peril_areas):
            self._peril_areas = self._stream = None
            kwargs['properties'] = RTreeIndexProperty(**props)
            super(self.__class__, self).__init__(*args, **kwargs)
        elif idx_fp:
            self._peril_areas = self._stream = None
            _idx_fp = idx_fp
            if not os.path.isabs(_idx_fp):
                _idx_fp = os.path.abspath(_idx_fp)

            idx_ext = props.get('idx_extension') or 'idx'
            dat_ext = props.get('dat_extension') or 'dat'

            if not (os.path.exists('{}.{}'.format(_idx_fp, idx_ext)) or os.path.exists('{}.{}'.format(_idx_fp, dat_ext))):
                kwargs['properties'] = RTreeIndexProperty(**props)

            super(self.__class__, self).__init__(_idx_fp, *args, **kwargs)
        else:
            self._peril_areas = OrderedDict({
                pa.id: pa for pa in (peril_areas if peril_areas else self._get_peril_areas(areas))
            })
            self._stream = self._generate_index_entries(
                ((paid, pa.bounds) for paid, pa in self._peril_areas.items()),
                objects=((paid, pa.bounds, pa.coordinates) for paid, pa in self._peril_areas.items())
            )
            kwargs['properties'] = RTreeIndexProperty(**props)
            super(self.__class__, self).__init__(self._stream, *args, **kwargs)

    def dumps(self, obj):
        return pickle.dumps(obj, protocol=self.protocol)

    def loads(self, data):
        return pickle.loads(data)

    def _get_peril_areas(self, areas):
        for peril_id, coverage_type, peril_area_id, coordinates, other_props in areas:
            yield PerilArea(coordinates, peril_id=peril_id, coverage_type=coverage_type, peril_area_id=peril_area_id, **other_props)

    def _generate_index_entries(self, items, objects=None):
        if objects:
            for (key, poly_bounds), obj in zip(items, objects):
                yield key, poly_bounds, obj
        else:
            for key, poly_bounds in items:
                yield key, poly_bounds, None

    @property
    def protocol(self):
        return self._protocol

    @property
    def peril_areas(self):
        return self._peril_areas

    @property
    def stream(self):
        if self._peril_areas:
            self._stream = self._generate_index_entries(self._peril_areas)
            return self._stream
        return None

    @classmethod
    def create_from_peril_areas_file(
        cls,
        src_fp=None,
        src_type='csv',
        peril_id_col='peril_id',
        coverage_type_col='coverage_type',
        peril_area_id_col='area_peril_id',
        non_na_cols=('peril_id', 'coverage_type', 'area_peril_id',),
        col_dtypes={'peril_id': int, 'coverage_type': int, 'area_peril_id': int},
        sort_cols=['area_peril_id'],
        area_poly_coords_cols={},
        area_poly_coords_seq_start_idx=1,
        area_reg_poly_radius=0.00166,
        static_props={},
        index_fp=None,
        index_props=copy.deepcopy(DEFAULT_RTREE_INDEX_PROPS)
    ):
        if not src_fp:
            raise OasisException(
                'An areas source CSV or JSON file path must be provided'
            )

        _src_fp = src_fp
        if not os.path.isabs(_src_fp):
            _src_fp = os.path.abspath(_src_fp)

        _non_na_cols = set(non_na_cols)

        _peril_id_col = peril_id_col.lower()
        _coverage_type_col = coverage_type_col.lower()
        _peril_area_id_col = peril_area_id_col.lower()

        if not set(_non_na_cols).intersection([_peril_id_col, _coverage_type_col, _peril_area_id_col]):
            _non_na_cols = _non_na_cols.union({_peril_id_col, _coverage_type_col, _peril_area_id_col})

        for col in area_poly_coords_cols.values():
            if col not in _non_na_cols:
                _non_na_cols = _non_na_cols.union({col.lower()})

        _non_na_cols = tuple(_non_na_cols)

        _sort_cols = [col.lower() for col in sort_cols]

        areas_df = get_dataframe(
            src_fp=_src_fp,
            src_type=src_type,
            non_na_cols=_non_na_cols,
            col_dtypes=col_dtypes,
            sort_cols=(_sort_cols or [_peril_area_id_col])
        )

        coords_cols = area_poly_coords_cols

        seq_start = area_poly_coords_seq_start_idx

        len_seq = sum(1 if re.match(r'x(\d+)?', k) else 0 for k in coords_cols.keys())

        peril_areas = cls()._get_peril_areas(
            (
                ar[_peril_id_col],
                ar[_coverage_type_col],
                ar[_peril_area_id_col],
                tuple(
                    (ar.get(coords_cols['x{}'.format(i)].lower()) or 0, ar.get(coords_cols['y{}'.format(i)].lower()) or 0)
                    for i in range(seq_start, len_seq + 1)
                ),
                static_props
            ) for _, ar in areas_df.iterrows()
        )

        _index_fp = index_fp
        if not _index_fp:
            raise OasisException('No output file index path provided')

        if not os.path.isabs(_index_fp):
            _index_fp = os.path.abspath(_index_fp)

        try:
            return cls().save(
                _index_fp,
                peril_areas=peril_areas,
                index_props=index_props
            )
        except OasisException:
            raise

    def save(
        self,
        index_fp,
        peril_areas=None,
        index_props=DEFAULT_RTREE_INDEX_PROPS
    ):
        _index_fp = index_fp

        if not os.path.isabs(_index_fp):
            _index_fp = os.path.abspath(_index_fp)

        if os.path.exists(_index_fp):
            os.remove(_index_fp)

        class myindex(RTreeIndex):
            def __init__(self, *args, **kwargs):
                self.protocol = pickle.HIGHEST_PROTOCOL
                super(self.__class__, self).__init__(*args, **kwargs)

            def dumps(self, obj):
                return pickle.dumps(obj, protocol=self.protocol)

            def loads(self, obj):
                return pickle.loads(obj)

        try:
            index = myindex(_index_fp, properties=RTreeIndexProperty(**index_props))

            _peril_areas = self._peril_areas or peril_areas

            if not _peril_areas:
                raise OasisException(
                    'No peril areas found in instance or in arguments - '
                    'this is required to write the index to file'
                )

            peril_areas_seq = None

            if (isinstance(peril_areas, list) or isinstance(peril_areas, tuple)):
                peril_areas_seq = (pa for pa in peril_areas)
            elif isinstance(peril_areas, types.GeneratorType):
                peril_areas_seq = peril_areas
            elif (isinstance(peril_areas, dict)):
                peril_areas_seq = peril_areas.values()

            for pa in peril_areas_seq:
                index.insert(pa.id, pa.bounds, obj=(pa.peril_id, pa.coverage_type, pa.id, pa.bounds, pa.coordinates))

            index.close()
        except (IOError, OSError, RTreeError):
            raise

        return _index_fp
